// implementation of declaration

#include <sdkddkver.h>
#include <afxwin.h>
#include "resource.h"
#include "editor.h"

BEGIN_MESSAGE_MAP(editor, CEdit)
    ON_WM_CONTEXTMENU()
    ON_WM_GETDLGCODE()
    ON_WM_SETFOCUS()
    ON_WM_SIZE()
END_MESSAGE_MAP()

editor::editor()
{
}

editor::~editor()
{
}

// it is called automatically
void editor::PreSubclassWindow()
{
    CEdit::PreSubclassWindow();

    // cursor of base of class is IDC_IBEAM
    // let us change it with our solid-beam

    // it can affect other windows of class
    // but it is more easy than OnSetCursor

    HCURSOR cursor = AfxGetApp()->LoadCursor(IDC_OBJECT_BEAM);
    if (cursor) SetClassLong(m_hWnd,GCL_HCURSOR,(long)cursor);
}

void editor::OnContextMenu(CWnd * window, CPoint point)
{
    // input may be WM_KEYUP (VK_APPS)
    // input may be WM_KEYDOWN (VK_SHIFT then VK_F10)
    // input may be WM_NCRBUTTONUP
    // input may be WM_RBUTTONUP

    // base of class may show copy and paste in its menu
    // but we want only copy, let us create special menu

    // we must pay attention to areas of bars of scrolls
    // because base of class may show side-menu in there

    if (window)
    {
     if (window->m_hWnd)
     {
      if (window->m_hWnd == m_hWnd)
      {
       CRect rectangle;
       GetClientRect(rectangle);
       ClientToScreen(rectangle);

       if ((point.x < 0) || (point.y < 0))
       {
        point = rectangle.TopLeft();
        ASSERT(rectangle.PtInRect(point));
       }

       if (rectangle.PtInRect(point))
       {
        CMenu menu;
        menu.CreatePopupMenu();
        menu.AppendMenu(MF_STRING,ID_EDIT_COPY,"Copy");
        menu.AppendMenu(MF_STRING,ID_EDIT_SELECT_ALL,"Select All");

        int head, tail; GetSel(head,tail);
        if (head == tail) menu.EnableMenuItem(ID_EDIT_COPY,MF_DISABLED|MF_GRAYED);
        if ((tail - head) == GetWindowTextLength()) menu.EnableMenuItem(ID_EDIT_SELECT_ALL,MF_DISABLED|MF_GRAYED);

        int result = menu.TrackPopupMenu(TPM_NONOTIFY|TPM_RETURNCMD,point.x,point.y,this);
        if (result == ID_EDIT_COPY) Copy();
        if (result == ID_EDIT_SELECT_ALL) SetSel(0,-1,TRUE);

        return;
       }

       // now let us leave
      }
     }
    }

    // in above, we changed the parameters
    // but, the change does not reach here

    CEdit::OnContextMenu(window,point);
}

UINT editor::OnGetDlgCode()
{
    unsigned int code = CEdit::OnGetDlgCode();

    // when window is first in order of tab
    // initialize of dialog does select all
    // key of tab in dialog does select all
    // let us prevent these because of want

    // avoid intermediate value
    code |= DLGC_HASSETSEL; // let us apply or-with-xor
    code ^= DLGC_HASSETSEL; // in place of and-with-not

    return code;
}

void editor::OnSetFocus(CWnd * older)
{
    CEdit::OnSetFocus(older);

    // base of class showed line-caret
    // and now let us show solid-caret

    // learn
    CDC context;
    context.CreateCompatibleDC(NULL);
    context.SetMapMode(MM_TEXT);
    CFont * font = GetFont();
    font = context.SelectObject(font);
    TEXTMETRIC metric = {};
    context.GetTextMetrics(&metric);
    font = context.SelectObject(font);
    context.DeleteDC();

    // check
    ASSERT(metric.tmPitchAndFamily == 0);

    // it does nope the old caret
    CreateSolidCaret(metric.tmAveCharWidth,metric.tmHeight);
    ShowCaret();

    // OnKillFocus of base of class does nope the caret
}

void editor::OnSize(unsigned int type, int width, int height)
{
    CEdit::OnSize(type,width,height);

    // for show good, let us put space to round of content

    // here does also fix the solid-caret on bar-of-scroll
    // moving solid-caret to end-of-line in most-long-line

    // EM_GETRECT may return approximate, already not need
    // we not need SM_CXBORDER, SM_CXVSCROLL, and so forth

    CRect rectangle;

    // default
    rectangle.top = 0;
    rectangle.left = 0;
    rectangle.right = width;
    rectangle.bottom = height;

    // deflate
    rectangle.top += 8;
    rectangle.left += 8;
    rectangle.right -= 8;
    rectangle.bottom -= 8;

    // arrange
    if (rectangle.top > rectangle.bottom) rectangle.SetRectEmpty();
    if (rectangle.left > rectangle.right) rectangle.SetRectEmpty();

    SetMargins(0,0);
    SetRect(rectangle);
}
