// implementation of declaration

// declared functions must be called from same thread
// and remained functions must not be called directly

// most of functions do not return results
// caller does understand from malfunction

// because of calls from threads at same time
// as much as we can, we included few headers

#include <sdkddkver.h>
#include <mfplay.h>
#include "neat.h"
#include "hook.h"

// nope in header
// internal usage
void hook_take();
void hook_give();

// our static data
data_of_hook hook_data;

// it must be called after all
// variable is not need for it
// usable as function().member
const data_of_hook & hook_info()
{
    return hook_data;
}

// it must be called from section
void hook_sort()
{
int index1;
int index2;
entry_of_hook temp;

    for (index1 = 0; index1 < RTL_NUMBER_OF(hook_data.entries) - 1; index1++)
    {
     if (hook_data.entries[index1].classy == GUID_NULL)
     {
      for (index2 = index1 + 1; index2 < RTL_NUMBER_OF(hook_data.entries); index2++)
      {
       if (hook_data.entries[index2].classy != GUID_NULL)
       {
        temp = hook_data.entries[index1];
        hook_data.entries[index1] = hook_data.entries[index2];
        hook_data.entries[index2] = temp;
        break;
       }
      }
      if (index2 >= RTL_NUMBER_OF(hook_data.entries)) break; // finish quick
     }
    }
}

// it must be called from section
void hook_free()
{
int index;
int count;

    for (index = 0; index < RTL_NUMBER_OF(hook_data.entries); index++)
    {
     if (hook_data.entries[index].classy == GUID_NULL) continue;
     if (hook_data.entries[index].object == NULL) continue;

     // temporary assign is to nullify nearby
     count = hook_data.entries[index].object->Release();
     hook_data.entries[index].object = NULL;

     if (count) continue; // caller has kept it

     zeroize(hook_data.entries[index]);
    }
}

// it must be called from section
void hook_quiz()
{
int index;
IMFAttributes * store;

    for (index = 0; index < RTL_NUMBER_OF(hook_data.entries); index++)
    {
     if (hook_data.entries[index].classy != GUID_NULL)
     {
      if (hook_data.entries[index].object)
      {
       if (hook_data.entries[index].demand == IID_IMFTransform)
       {
        if (SUCCEEDED(((IMFTransform *)hook_data.entries[index].object)->GetAttributes(&store)))
        {
         store->GetUINT32(MF_SA_D3D_AWARE,&hook_data.entries[index].harder);
         store->GetStringLength(MFT_ENUM_HARDWARE_URL_Attribute,&hook_data.entries[index].selfer);
         store->Release(); store = NULL;
        }
       }
      }
     }
    }
}

// it must be called from section
void hook_save(REFCLSID classy, REFIID demand, LPCVOID object)
{
int index;

    // Enhanced Video Renderer does not come here

    // Sample Copier Transform does not come here

    // Streaming Audio Renderer does not come here

    // Video Mixer does come here

    // Video Presenter does come here

    // when extension and content are different
    // caller does scan the handlers one-by-one

    if (demand != IID_IMFSchemeHandler)
    if (demand != IID_IMFByteStreamHandler)
    if (demand != IID_IMFTransform)
    if (demand != IID_IMFVideoPresenter)
    return; // not relevant to topology

    for (index = 0; index < RTL_NUMBER_OF(hook_data.entries); index++)
    {
     if (hook_data.entries[index].classy == GUID_NULL)
     {
      hook_data.entries[index].classy = classy;
      hook_data.entries[index].demand = demand;
      hook_data.entries[index].object = NULL;
      hook_data.entries[index].harder = 0;
      hook_data.entries[index].selfer = 0;

      if (demand == IID_IMFTransform)
      {
       // it does inherit from other
       hook_data.entries[index].object = (LPUNKNOWN)object;
       hook_data.entries[index].object->AddRef();
      }

      if (demand == IID_IMFVideoPresenter)
      {
       // it does inherit from other
       hook_data.entries[index].object = (LPUNKNOWN)object;
       hook_data.entries[index].object->AddRef();
      }

      break;
     }

     if (demand == IID_IMFSchemeHandler)
     {
      if (demand == hook_data.entries[index].demand)
      {
       // pass onto it
       hook_data.entries[index].classy = classy;
       break;
      }
     }

     if (demand == IID_IMFByteStreamHandler)
     {
      if (demand == hook_data.entries[index].demand)
      {
       // pass onto it
       hook_data.entries[index].classy = classy;
       break;
      }
     }

     // continue of loop
    }

    if (index >= RTL_NUMBER_OF(hook_data.entries)) ExitProcess(ERROR_PROCESS_ABORTED);
}

// it is defined such actual function
// (threads may call it at same time)
HRESULT STDAPICALLTYPE hook_ours(REFCLSID classy, LPUNKNOWN outer, DWORD context, REFIID demand, PVOID * object)
{
HRESULT result;

    EnterCriticalSection(&hook_data.section);

    hook_give(); // put its instructions to work them

    // call the actual function
    result = CoCreateInstance(classy,outer,context,demand,object);
    if (SUCCEEDED(result)) hook_save(classy,demand,*object);

    hook_take(); // put our instructions to come here

    LeaveCriticalSection(&hook_data.section);

    return result;
}

// it must be called from section
// (excess call does not problem)
// also actual function may crash
// (in other thread at same time)
// (because its code does change)
void hook_give()
{
ULONG old;

    if (hook_data.library)
    {
     if (hook_data.address)
     {
      if (VirtualProtect(hook_data.address,sizeof(hook_data.content),PAGE_EXECUTE_READWRITE,&old))
      {
       CopyMemory(hook_data.address,hook_data.content,sizeof(hook_data.content));
       VirtualProtect(hook_data.address,sizeof(hook_data.content),old,&old);
       FlushInstructionCache(GetCurrentProcess(),hook_data.address,sizeof(hook_data.content));
      }
     }
    }
}

// it must be called from section
// (excess call does not problem)
// also actual function may crash
// (in other thread at same time)
// (because its code does change)
void hook_take()
{
ULONG old;
ULONG index;

    if (hook_data.library)
    {
     if (hook_data.address)
     {
      if (VirtualProtect(hook_data.address,sizeof(hook_data.content),PAGE_EXECUTE_READWRITE,&old))
      {
       index = 0;

       // instruction of jump (in mods of 32 with 64)
       *(PUCHAR)((ULONG)hook_data.address + index) = 0xE9;
       index += sizeof(UCHAR);

       // displacement of jump (relative 32 bits)
       *(PULONG)((ULONG)hook_data.address + index) = (ULONG)hook_ours - (ULONG)hook_data.address - index - sizeof(ULONG);
       index += sizeof(ULONG);

       if (index > sizeof(hook_data.content)) ExitProcess(ERROR_PROCESS_ABORTED);

       VirtualProtect(hook_data.address,sizeof(hook_data.content),old,&old);
       FlushInstructionCache(GetCurrentProcess(),hook_data.address,sizeof(hook_data.content));
      }
     }
    }
}

// it must be called at end
void hook_exit()
{
    EnterCriticalSection(&hook_data.section);

    if (hook_data.library)
    {
     hook_give();
     hook_quiz();
     hook_free();
     hook_sort();
     FreeLibrary(hook_data.library);
     hook_data.library = NULL;
     hook_data.address = NULL;
    }

    LeaveCriticalSection(&hook_data.section);
    DeleteCriticalSection(&hook_data.section);

    // still, other threads may attempt to use
    // for this problem, caller is responsible
}

// it must be called at begin
void hook_init()
{
    zeroize(hook_data);
    InitializeCriticalSection(&hook_data.section);
    EnterCriticalSection(&hook_data.section);

    hook_data.library = LoadLibrary("ole32.dll");
    if (hook_data.library)
    {
     hook_data.address = GetProcAddress(hook_data.library,"CoCreateInstance");
     if (hook_data.address)
     {
      CopyMemory(hook_data.content,hook_data.address,sizeof(hook_data.content));
      hook_take();
     }
    }

    LeaveCriticalSection(&hook_data.section);

    // till end of all of work
    // let library remain open
}
