// implementation of declaration

// because of calls from threads at same time
// as much as we can, we included few headers

#include <sdkddkver.h>
#include <mfapi.h>
#include <mfplay.h>
#include "neat.h"
#include "qualifier.h"

// our static data
data_of_quality quality_data;

IMFTopology * read_the_topology()
{
    // we not used the count of reference of it
    // else object of lock is need at this time

    // pending because of threads
    return quality_data.topology;
}

long read_then_zeroize_the_cleans_of_audio()
{
    // lock is because of threads
    return InterlockedExchange(&quality_data.cleans_of_audio,0);
}

long read_then_zeroize_the_cleans_of_video()
{
    // lock is because of threads
    return InterlockedExchange(&quality_data.cleans_of_video,0);
}

long read_then_zeroize_the_passes_of_audio()
{
    // lock is because of threads
    return InterlockedExchange(&quality_data.passes_of_audio,0);
}

long read_then_zeroize_the_passes_of_video()
{
    // lock is because of threads
    return InterlockedExchange(&quality_data.passes_of_video,0);
}

qualifier::qualifier(IMFQualityManager * object)
{
    m_count = 1; // after create, increment is not need
    m_object = object;
    zeroize(quality_data);
}

qualifier::~qualifier()
{
    if (m_object)
    {
     m_object->Release(); m_object = NULL;
    }

    zeroize(quality_data);
}

ULONG STDMETHODCALLTYPE qualifier::AddRef()
{
    // lock is because of threads
    return InterlockedIncrement(&m_count);
}

ULONG STDMETHODCALLTYPE qualifier::Release()
{
    // lock is because of threads
    int count = InterlockedDecrement(&m_count);
    if (count == 0) delete this; return count;
}

HRESULT STDMETHODCALLTYPE qualifier::QueryInterface(REFIID demand, void ** object)
{
    if (object == NULL) return E_INVALIDARG;

    *object = NULL;

    if (demand == IID_IUnknown)
    {
     // check of type, safe than cast
     IUnknown * pointer = this;
     pointer->AddRef(); *object = pointer;
     return S_OK;
    }

    if (demand == IID_IMFQualityManager)
    {
     // check of type, safe than cast
     IMFQualityManager * pointer = this;
     pointer->AddRef(); *object = pointer;
     return S_OK;
    }

    return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE qualifier::NotifyPresentationClock(IMFPresentationClock * clock)
{
    return m_object->NotifyPresentationClock(clock);
}

HRESULT STDMETHODCALLTYPE qualifier::NotifyProcessInput(IMFTopologyNode * node, long index, IMFSample * sample)
{
    // node of output does come here
    // node of source does not come here
    // node of transform does come here

    notify_input(node,index,sample); // we can change the parameters at there

    return m_object->NotifyProcessInput(node,index,sample);
}

HRESULT STDMETHODCALLTYPE qualifier::NotifyProcessOutput(IMFTopologyNode * node, long index, IMFSample * sample)
{
    // node of output does not come here
    // node of source does not come here
    // node of transform does come here

    return m_object->NotifyProcessOutput(node,index,sample);
}

HRESULT STDMETHODCALLTYPE qualifier::NotifyQualityEvent(IUnknown * object, IMFMediaEvent * event)
{
    return m_object->NotifyQualityEvent(object,event);
}

HRESULT STDMETHODCALLTYPE qualifier::NotifyTopology(IMFTopology * topology)
{
    // lock is because of threads
    InterlockedExchangePointer(&quality_data.topology,topology);

    return m_object->NotifyTopology(topology);
}

HRESULT STDMETHODCALLTYPE qualifier::Shutdown()
{
    // lock is because of threads
    InterlockedExchangePointer(&quality_data.topology,NULL);

    return m_object->Shutdown();
}

void qualifier::notify_input(IMFTopologyNode * node, DWORD index, IMFSample * sample)
{
HRESULT result;
MF_TOPOLOGY_TYPE type;
GUID major;
IMFTopologyNode * other;
IMFStreamDescriptor * descriptor;
IMFMediaTypeHandler * handler;

    // to avoid latency in flow
    // here must return quickly

    // to eliminate the repeats of notify
    // let us research the node of source
    // we know that it does not come here
    // also flag of clean is actual in it

    // we query MF_TOPOLOGY_SOURCESTREAM_NODE
    // although MF_TOPONODE_STREAM_DESCRIPTOR
    // however it does quicken here strangely

    if (node == NULL) return;
    if (sample == NULL) return;

    result = node->GetInput(index,&other,&index);
    if (SUCCEEDED(result))
    {
     result = other->GetNodeType(&type);
     if (SUCCEEDED(result))
     {
      if (type == MF_TOPOLOGY_SOURCESTREAM_NODE)
      {
       result = other->GetUnknown(MF_TOPONODE_STREAM_DESCRIPTOR,IID_PPV_ARGS(&descriptor));
       if (SUCCEEDED(result))
       {
        result = descriptor->GetMediaTypeHandler(&handler);
        if (SUCCEEDED(result))
        {
         result = handler->GetMajorType(&major);
         if (SUCCEEDED(result))
         {
          // lock is because of threads
          if (major == MFMediaType_Audio) InterlockedIncrement(&quality_data.passes_of_audio);
          if (major == MFMediaType_Video) InterlockedIncrement(&quality_data.passes_of_video);
          if (major == MFMediaType_Audio) if (MFGetAttributeUINT32(sample,MFSampleExtension_CleanPoint,FALSE)) InterlockedIncrement(&quality_data.cleans_of_audio);
          if (major == MFMediaType_Video) if (MFGetAttributeUINT32(sample,MFSampleExtension_CleanPoint,FALSE)) InterlockedIncrement(&quality_data.cleans_of_video);
         }
         handler->Release(); handler = NULL;
        }
        descriptor->Release(); descriptor = NULL;
       }
       type = MF_TOPOLOGY_MAX; // hint of block
      }
      type = MF_TOPOLOGY_MAX; // hint of block
     }
     other->Release(); other = NULL;
    }
}
