// implementation of declaration

#include <sdkddkver.h>
#include <afxwin.h>
#include <mfapi.h>
#include <mfplay.h>
#include "lockee.h"
#include "locker.h"
#include "backer.h"
#include "common.h"
#include "hook.h"
#include "hook_2.h"
#include "media.h"
#include "viewer.h"

BEGIN_MESSAGE_MAP(viewer, CWnd)
    ON_WM_PAINT()
    ON_WM_SIZE()
    ON_WM_LBUTTONDOWN()
    ON_WM_LBUTTONUP()
    ON_WM_MOUSEMOVE()
END_MESSAGE_MAP()

viewer::viewer()
{
    m_back = NULL;
    m_play = NULL;
    m_item = NULL;

    m_state = MFP_MEDIAPLAYER_STATE_EMPTY;

    m_audio_stream = FALSE;
    m_audio_select = FALSE;
    m_video_stream = FALSE;
    m_video_select = FALSE;

    m_drag = false;

    m_point.x = 0;
    m_point.y = 0;
}

viewer::~viewer()
{
    conclude();
}

// from time to time, it is called from some places
void viewer::conclude()
{
    if (m_play) m_play->Shutdown(); // it is synchronous

    // because of the shutdown from now on
    // it is safe to release of the remain
    // order will be reverse of the create

    if (m_item)
    {
     // may be player has increased it
     int count = m_item->Release(); m_item = NULL;
     ASSERT(count <= 1); // not must but most
    }

    if (m_play)
    {
     // destroy of player does call of MFShutdown
     int count = m_play->Release(); m_play = NULL;
     ASSERT(count == 0); // not must but most
    }

    if (m_back)
    {
     int count = m_back->Release(); m_back = NULL;
     ASSERT(count == 0); // not must but most
    }

    m_state = MFP_MEDIAPLAYER_STATE_EMPTY;

    m_audio_stream = FALSE;
    m_audio_select = FALSE;
    m_video_stream = FALSE;
    m_video_select = FALSE;

    m_drag = false;

    m_point.x = 0;
    m_point.y = 0;
}

// from time to time, it is called from some places
LRESULT viewer::commence(PCSTR name, const bool audio, const bool video)
{
long result; // for HRESULT with LRESULT

    if (m_back) return ERROR_ALREADY_EXISTS;
    if (m_play) return ERROR_ALREADY_EXISTS;
    if (m_item) return ERROR_ALREADY_EXISTS;

    hook_init();
    hook_init_2();

    m_back = new backer; result = m_back ? S_OK : E_FAIL;

    if (SUCCEEDED(result))
    {
     // create of player does call of MFStartup
     result = MFPCreateMediaPlayer(NULL,FALSE,MFP_OPTION_FREE_THREADED_CALLBACK,m_back,m_hWnd,&m_play);
    }

    if (SUCCEEDED(result))
    {
     result = m_play->CreateMediaItemFromURL(CStringW(name),TRUE,0,&m_item);
    }

    if (SUCCEEDED(result))
    {
     if (audio == false) close_stream(m_item,MFMediaType_Audio);
    }

    if (SUCCEEDED(result))
    {
     if (video == false) close_stream(m_item,MFMediaType_Video);
    }

    if (SUCCEEDED(result))
    {
     result = m_item->HasAudio(&m_audio_stream,&m_audio_select);
    }

    if (SUCCEEDED(result))
    {
     result = m_item->HasVideo(&m_video_stream,&m_video_select);
    }

    if (SUCCEEDED(result))
    {
     result = m_play->SetMediaItem(m_item);
    }

    // last of operations was asynchronous
    // its asynchronous result is not need

    if (SUCCEEDED(result))
    {
     while (m_back->status().eEventType != MFP_EVENT_TYPE_MEDIAITEM_SET) Sleep(100);
    }

    // we understood that topology completed
    // we used back of call to understand it
    // IMFPMediaPlayer is a way too for this
    // but a stream must reach to a renderer
    // else it does return fail continuously
    // this method can not be useful in here

    if (SUCCEEDED(result))
    {
     m_state = MFP_MEDIAPLAYER_STATE_STOPPED;
    }

    if (SUCCEEDED(result))
    {
     result = ERROR_SUCCESS;
    }

    hook_exit_2();
    hook_exit();

    if (result) conclude();

    return result;
}

BOOL viewer::PreCreateWindow(CREATESTRUCT & data)
{
    if (CWnd::PreCreateWindow(data) == FALSE) return FALSE;

    data.style = WS_CHILD | WS_VISIBLE;
    data.dwExStyle = 0;
    data.lpszClass = AfxRegisterWndClass(0,LoadCursor(NULL,IDC_SIZEALL /* also known as move */),CreateSolidBrush(0),NULL);

    return TRUE;
}

void viewer::OnPaint()
{
    CWnd::OnPaint();

    if (m_play) m_play->UpdateVideo();
}

void viewer::OnSize(unsigned int type, int width, int height)
{
    CWnd::OnSize(type,width,height);

    if (m_play) m_play->UpdateVideo();
}

void viewer::OnLButtonDown(unsigned int flags, CPoint point)
{
    m_drag = true;
    m_point = point;

    CWnd::OnLButtonDown(flags,point);
}

void viewer::OnLButtonUp(unsigned int flags, CPoint point)
{
    m_drag = false;
    m_point = point;

    CWnd::OnLButtonUp(flags,point);
}

void viewer::OnMouseMove(unsigned int flags, CPoint point)
{
RECT client;
SIZE video;
SIZE target;
SIZE step_of_target;
size_as_float step_of_source;
MFVideoNormalizedRect source;

    if (flags == MK_LBUTTON) // let us not allow other flags
    if (m_drag)
    if (m_play)
    if (m_state == MFP_MEDIAPLAYER_STATE_PAUSED) // drag of user does lag in state of play
    if (m_video_stream)
    if (m_video_select)
    if (SUCCEEDED(m_play->GetNativeVideoSize(&video,NULL)))
    if (SUCCEEDED(m_play->GetVideoSourceRect(&source)))
    {
     // about overflow, underflow, and precision
     // type of float is enough in small numbers

     // let us pay attention to divide by zero

     // scope of video
     GetClientRect(&client);

     // values can be zero
     target.cx = client.right - client.left; // left and top are zero
     target.cy = client.bottom - client.top; // but for compatibility

     // calculation of target is approximate
     // because we can not reach to renderer
     if ((target.cx * video.cy) < (target.cy * video.cx))
     {
      target.cy = (target.cx * video.cy) / video.cx;
     }
     else
     {
      target.cx = (target.cy * video.cx) / video.cy;
     }

     // avoid divide by zero
     if (target.cx && target.cy)
     {
      // values can be zero
      step_of_target.cx = point.x - m_point.x;
      step_of_target.cy = point.y - m_point.y;

      step_of_source.cx = source.right - source.left;
      step_of_source.cx *= step_of_target.cx;
      step_of_source.cx /= target.cx;

      step_of_source.cy = source.bottom - source.top;
      step_of_source.cy *= step_of_target.cy;
      step_of_source.cy /= target.cy;

      source.top -= step_of_source.cy;
      source.left -= step_of_source.cx;
      source.right -= step_of_source.cx;
      source.bottom -= step_of_source.cy;
     }

     normalize_rectangle(source);
     m_play->SetVideoSourceRect(&source);
    }

    // m_drag must not be updated in the function
    // else drag may begin from outside of window
    m_point = point;

    CWnd::OnMouseMove(flags,point);
}
